@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico.
-   Copyright (C) 2008-2018 Sergey Poznyakoff
+   Copyright (C) 2008-2021 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -83,7 +83,7 @@ static struct dico_strategy regex_strat = {
 };
 
 void
-register_regex()
+register_regex(void)
 {
     dico_strategy_add(&re_strat);
     dico_strategy_add(&regex_strat);
